import numpy as np
import matplotlib.pyplot as plt

from .__version__ import __version__
from .__version__ import __version_info__
from .__version__ import __author__

from .classes import TimeSerie
from .classes import BooleanTimeSerie
from .classes import StateChangeArray
from .classes import BooleanStateChangeArray
from .classes import Report
from .classes import Event


plt.ion()
